@@ -66,6 +66,9 @@ void vk_cmd_dep(struct vk_cmd *cmd, VkPipelineStageFlags2 stage, pl_vulkan_sem d
 // after the given stage completes.
 void vk_cmd_sig(struct vk_cmd *cmd, VkPipelineStageFlags2 stage, pl_vulkan_sem sig);
 
+// Compatibility wrappers for vkCmdPipelineBarrier2 (works with pre-1.3)
+void vk_cmd_barrier(struct vk_cmd *cmd, const VkDependencyInfo *info);
+
 // Synchronization scope
 struct vk_sync_scope {
     pl_vulkan_sem sync;         // semaphore of last access
@@ -84,12 +87,13 @@ struct vk_sem {
 //
 // Returns a struct describing the previous access to a resource. A pipeline
 // barrier is only required if the previous access scope is nonzero.
-struct vk_sync_scope vk_sem_barrier(struct vk_ctx *vk, struct vk_cmd *cmd,
-                                    struct vk_sem *sem, VkPipelineStageFlags2 stage,
+struct vk_sync_scope vk_sem_barrier(struct vk_cmd *cmd, struct vk_sem *sem,
+                                    VkPipelineStageFlags2 stage,
                                     VkAccessFlags2 access, bool is_trans);
 
 // Command pool / queue family hybrid abstraction
 struct vk_cmdpool {
+    struct vk_ctx *vk;
     VkQueueFamilyProperties props;
     int qf; // queue family index
     VkCommandPool pool;
@@ -106,16 +110,15 @@ struct vk_cmdpool {
 struct vk_cmdpool *vk_cmdpool_create(struct vk_ctx *vk, int qf, int qnum,
                                      VkQueueFamilyProperties props);
 
-void vk_cmdpool_destroy(struct vk_ctx *vk, struct vk_cmdpool *pool);
+void vk_cmdpool_destroy(struct vk_cmdpool *pool);
 
 // Fetch a command buffer from a command pool and begin recording to it.
 // Returns NULL on failure.
-struct vk_cmd *vk_cmd_begin(struct vk_ctx *vk, struct vk_cmdpool *pool,
-                            pl_debug_tag debug_tag);
+struct vk_cmd *vk_cmd_begin(struct vk_cmdpool *pool, pl_debug_tag debug_tag);
 
 // Finish recording a command buffer and submit it for execution. This function
 // takes over ownership of **cmd, and sets *cmd to NULL in doing so.
-bool vk_cmd_submit(struct vk_ctx *vk, struct vk_cmd **cmd);
+bool vk_cmd_submit(struct vk_cmd **cmd);
 
 // Block until some commands complete executing. This is the only function that
 // actually processes the callbacks. Will wait at most `timeout` nanoseconds
