@@ -39,7 +39,7 @@ void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
 
     struct vk_sync_scope last;
     bool is_trans = layout != tex_vk->layout, is_xfer = qf != tex_vk->qf;
-    last = vk_sem_barrier(vk, cmd, &tex_vk->sem, stage, access, is_trans || is_xfer);
+    last = vk_sem_barrier(cmd, &tex_vk->sem, stage, access, is_trans || is_xfer);
 
     VkImageMemoryBarrier2 barr = {
         .sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER_2,
@@ -65,7 +65,7 @@ void vk_tex_barrier(pl_gpu gpu, struct vk_cmd *cmd, pl_tex tex,
     }
 
     if (last.access || is_trans || is_xfer) {
-        vk->CmdPipelineBarrier2KHR(cmd->buf, &(VkDependencyInfo) {
+        vk_cmd_barrier(cmd, &(VkDependencyInfo) {
             .sType = VK_STRUCTURE_TYPE_DEPENDENCY_INFO,
             .imageMemoryBarrierCount = 1,
             .pImageMemoryBarriers = &barr,
@@ -96,6 +96,7 @@ static void vk_tex_destroy(pl_gpu gpu, struct pl_tex_t *tex)
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
 
     vk_sync_deref(gpu, tex_vk->ext_sync);
+    vk->DestroyFramebuffer(vk->dev, tex_vk->framebuffer, PL_VK_ALLOC);
     vk->DestroyImageView(vk->dev, tex_vk->view, PL_VK_ALLOC);
     for (int i = 0; i < tex_vk->num_planes; i++)
         vk_tex_deref(gpu, tex->planes[i]);
@@ -118,7 +119,7 @@ void vk_tex_deref(pl_gpu gpu, pl_tex tex)
 }
 
 
-// Initializes non-VkImage values like the image view and metadata
+// Initializes non-VkImage values like the image view, framebuffers, etc.
 static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
 {
     struct pl_vk *p = PL_PRIV(gpu);
@@ -144,6 +145,9 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
     if (params->format->emulated)
         tex_vk->transfer_queue = COMPUTE;
 
+    bool ret = false;
+    VkRenderPass dummyPass = VK_NULL_HANDLE;
+
     if (params->sampleable || params->renderable || params->storable) {
         static const VkImageViewType viewType[] = {
             [VK_IMAGE_TYPE_1D] = VK_IMAGE_VIEW_TYPE_1D,
@@ -167,10 +171,64 @@ static bool vk_init_image(pl_gpu gpu, pl_tex tex, pl_debug_tag debug_tag)
         PL_VK_NAME(IMAGE_VIEW, tex_vk->view, debug_tag);
     }
 
-    return true;
+    if (params->renderable) {
+        // Framebuffers need to be created against a specific render pass
+        // layout, so we need to temporarily create a skeleton/dummy render
+        // pass for vulkan to figure out the compatibility
+        VkRenderPassCreateInfo rinfo = {
+            .sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO,
+            .attachmentCount = 1,
+            .pAttachments = &(VkAttachmentDescription) {
+                .format = tex_vk->img_fmt,
+                .samples = VK_SAMPLE_COUNT_1_BIT,
+                .loadOp = VK_ATTACHMENT_LOAD_OP_DONT_CARE,
+                .storeOp = VK_ATTACHMENT_STORE_OP_STORE,
+                .initialLayout = VK_IMAGE_LAYOUT_UNDEFINED,
+                .finalLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
+            },
+            .subpassCount = 1,
+            .pSubpasses = &(VkSubpassDescription) {
+                .pipelineBindPoint = VK_PIPELINE_BIND_POINT_GRAPHICS,
+                .colorAttachmentCount = 1,
+                .pColorAttachments = &(VkAttachmentReference) {
+                    .attachment = 0,
+                    .layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL,
+                },
+            },
+        };
+
+        VK(vk->CreateRenderPass(vk->dev, &rinfo, PL_VK_ALLOC, &dummyPass));
+
+        VkFramebufferCreateInfo finfo = {
+            .sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO,
+            .renderPass = dummyPass,
+            .attachmentCount = 1,
+            .pAttachments = &tex_vk->view,
+            .width = tex->params.w,
+            .height = tex->params.h,
+            .layers = 1,
+        };
+
+        if (finfo.width > vk->props.limits.maxFramebufferWidth ||
+            finfo.height > vk->props.limits.maxFramebufferHeight)
+        {
+            PL_ERR(gpu, "Framebuffer of size %dx%d exceeds the maximum allowed "
+                   "dimensions: %dx%d", finfo.width, finfo.height,
+                   vk->props.limits.maxFramebufferWidth,
+                   vk->props.limits.maxFramebufferHeight);
+            goto error;
+        }
+
+        VK(vk->CreateFramebuffer(vk->dev, &finfo, PL_VK_ALLOC,
+                                 &tex_vk->framebuffer));
+        PL_VK_NAME(FRAMEBUFFER, tex_vk->framebuffer, debug_tag);
+    }
+
+    ret = true;
 
 error:
-    return false;
+    vk->DestroyRenderPass(vk->dev, dummyPass, PL_VK_ALLOC);
+    return ret;
 }
 
 pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
@@ -211,20 +269,6 @@ pl_tex vk_tex_create(pl_gpu gpu, const struct pl_tex_params *params)
             goto error;
         }
 
-        // Statically check to see if we'd even be able to upload it at all
-        // and refuse right away if not. In theory, uploading can still fail
-        // based on the size of pl_tex_transfer_params.row_pitch, but for now
-        // this should be enough.
-        uint64_t texels = params->w * PL_DEF(params->h, 1) * PL_DEF(params->d, 1) *
-                          fmt->num_components;
-
-        if (texels > gpu->limits.max_buffer_texels) {
-            PL_ERR(gpu, "Failed creating texture with emulated texture format: "
-                   "texture dimensions exceed maximum texel buffer size! Try "
-                   "again with a different (non-emulated) format?");
-            goto error;
-        }
-
         // Our format emulation requires storage image support. In order to
         // make a bunch of checks happy, just mark it off as storable (and also
         // enable VK_IMAGE_USAGE_STORAGE_BIT, which we do below)
@@ -816,6 +860,8 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     pl_tex tex = params->tex;
     pl_fmt fmt = tex->params.format;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+    struct pl_tex_transfer_params *slices = NULL;
+    int num_slices = 0;
 
     if (!params->buf)
         return pl_tex_upload_pbo(gpu, params);
@@ -823,74 +869,74 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     pl_buf buf = params->buf;
     struct pl_buf_vk *buf_vk = PL_PRIV(buf);
     pl_rect3d rc = params->rc;
-    size_t size = pl_tex_transfer_size(params);
-
-    size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
+    const size_t size = pl_tex_transfer_size(params);
+    const size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
     bool unaligned = buf_offset % fmt->texel_size;
     if (unaligned)
         PL_TRACE(gpu, "vk_tex_upload: unaligned transfer (slow path)");
 
     if (fmt->emulated || unaligned) {
 
-        // Copy the source data buffer into an intermediate buffer
-        struct pl_buf_params tbuf_params = {
-            .debug_tag = PL_DEBUG_TAG,
-            .memory_type = PL_BUF_MEM_DEVICE,
-            .format = tex_vk->texel_fmt,
-            .size = size,
-        };
+        // Create all slice buffers first, to early-fail if OOM, and to avoid
+        // blocking unnecessarily on waiting for these buffers to get read from
+        num_slices = pl_tex_transfer_slices(gpu, tex_vk->texel_fmt, params, &slices);
+        for (int i = 0; i < num_slices; i++) {
+            slices[i].buf = pl_buf_create(gpu, pl_buf_params(
+                .memory_type = PL_BUF_MEM_DEVICE,
+                .format      = tex_vk->texel_fmt,
+                .size        = pl_tex_transfer_size(&slices[i]),
+                .storable    = fmt->emulated,
+            ));
 
-        if (fmt->emulated) {
-            if (size <= gpu->limits.max_ubo_size) {
-                tbuf_params.uniform = true;
-            } else if (size <= gpu->limits.max_ssbo_size) {
-                tbuf_params.storable = true;
-            } else {
-                // TODO: Implement strided upload path if really necessary
-                PL_ERR(gpu, "Texel buffer size requirements exceed GPU "
-                       "capabilities, failed uploading!");
+            if (!slices[i].buf) {
+                PL_ERR(gpu, "Failed creating buffer for tex upload fallback!");
+                num_slices = i; // only clean up buffers up to here
                 goto error;
             }
         }
 
-        pl_buf tbuf = pl_buf_create(gpu, &tbuf_params);
-        if (!tbuf) {
-            PL_ERR(gpu, "Failed creating buffer for tex upload fallback!");
-            goto error;
-        }
-
-        struct vk_cmd *cmd = CMD_BEGIN_TIMED(tex_vk->transfer_queue, params->timer);
+        // All temporary buffers successfully created, begin copying source data
+        struct vk_cmd *cmd = CMD_BEGIN_TIMED(tex_vk->transfer_queue,
+                                             params->timer);
         if (!cmd)
             goto error;
 
-        struct pl_buf_vk *tbuf_vk = PL_PRIV(tbuf);
-        VkBufferCopy region = {
-            .srcOffset = buf_offset,
-            .dstOffset = tbuf_vk->mem.offset,
-            .size = size,
-        };
-
         vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
                        VK_ACCESS_2_TRANSFER_READ_BIT, params->buf_offset, size,
                        false);
-        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_2_COPY_BIT,
-                       VK_ACCESS_2_TRANSFER_WRITE_BIT, 0, size, false);
-        vk->CmdCopyBuffer(cmd->buf, buf_vk->mem.buf, tbuf_vk->mem.buf,
-                          1, &region);
+
+        for (int i = 0; i < num_slices; i++) {
+            pl_buf slice = slices[i].buf;
+            struct pl_buf_vk *slice_vk = PL_PRIV(slice);
+            vk_buf_barrier(gpu, cmd, slice, VK_PIPELINE_STAGE_2_COPY_BIT,
+                           VK_ACCESS_2_TRANSFER_WRITE_BIT, 0, slice->params.size,
+                           false);
+
+            vk->CmdCopyBuffer(cmd->buf, buf_vk->mem.buf, slice_vk->mem.buf, 1, &(VkBufferCopy) {
+                .srcOffset = buf_vk->mem.offset + slices[i].buf_offset,
+                .dstOffset = slice_vk->mem.offset,
+                .size      = slice->params.size,
+            });
+        }
 
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-        CMD_FINISH(&cmd);
+        bool ok = CMD_FINISH(&cmd);
 
-        struct pl_tex_transfer_params fixed = *params;
-        fixed.buf = tbuf;
-        fixed.buf_offset = 0;
+        // Finally, dispatch the (texel) upload asynchronously. We can fire
+        // the callback already at the completion of previous command because
+        // these temporary buffers already hold persistent copies of the data
+        for (int i = 0; i < num_slices; i++) {
+            if (ok) {
+                slices[i].buf_offset = 0;
+                ok = fmt->emulated ? pl_tex_upload_texel(gpu, &slices[i])
+                                   : pl_tex_upload(gpu, &slices[i]);
+            }
+            pl_buf_destroy(gpu, &slices[i].buf);
+        }
 
-        bool ok = fmt->emulated ? pl_tex_upload_texel(gpu, &fixed)
-                                : pl_tex_upload(gpu, &fixed);
-
-        pl_buf_destroy(gpu, &tbuf);
+        pl_free(slices);
         return ok;
 
     } else {
@@ -926,12 +972,15 @@ bool vk_tex_upload(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-        CMD_FINISH(&cmd);
+        return CMD_FINISH(&cmd);
     }
 
-    return true;
+    pl_unreachable();
 
 error:
+    for (int i = 0; i < num_slices; i++)
+        pl_buf_destroy(gpu, &slices[i].buf);
+    pl_free(slices);
     return false;
 }
 
@@ -942,6 +991,8 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     pl_tex tex = params->tex;
     pl_fmt fmt = tex->params.format;
     struct pl_tex_vk *tex_vk = PL_PRIV(tex);
+    struct pl_tex_transfer_params *slices = NULL;
+    int num_slices = 0;
 
     if (!params->buf)
         return pl_tex_download_pbo(gpu, params);
@@ -949,71 +1000,80 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
     pl_buf buf = params->buf;
     struct pl_buf_vk *buf_vk = PL_PRIV(buf);
     pl_rect3d rc = params->rc;
-    size_t size = pl_tex_transfer_size(params);
-
-    size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
+    const size_t size = pl_tex_transfer_size(params);
+    const size_t buf_offset = buf_vk->mem.offset + params->buf_offset;
     bool unaligned = buf_offset % fmt->texel_size;
     if (unaligned)
         PL_TRACE(gpu, "vk_tex_download: unaligned transfer (slow path)");
 
     if (fmt->emulated || unaligned) {
 
-        // Download into an intermediate buffer first
-        pl_buf tbuf = pl_buf_create(gpu, pl_buf_params(
-            .storable = fmt->emulated,
-            .size = size,
-            .memory_type = PL_BUF_MEM_DEVICE,
-            .format = tex_vk->texel_fmt,
-        ));
+        num_slices = pl_tex_transfer_slices(gpu, tex_vk->texel_fmt, params, &slices);
+        for (int i = 0; i < num_slices; i++) {
+            slices[i].buf = pl_buf_create(gpu, pl_buf_params(
+                .memory_type = PL_BUF_MEM_DEVICE,
+                .format      = tex_vk->texel_fmt,
+                .size        = pl_tex_transfer_size(&slices[i]),
+                .storable    = fmt->emulated,
+            ));
 
-        if (!tbuf) {
-            PL_ERR(gpu, "Failed creating buffer for tex download fallback!");
-            goto error;
+            if (!slices[i].buf) {
+                PL_ERR(gpu, "Failed creating buffer for tex download fallback!");
+                num_slices = i;
+                goto error;
+            }
         }
 
-        struct pl_tex_transfer_params fixed = *params;
-        fixed.buf = tbuf;
-        fixed.buf_offset = 0;
-
-        bool ok = fmt->emulated ? pl_tex_download_texel(gpu, &fixed)
-                                : pl_tex_download(gpu, &fixed);
-        if (!ok) {
-            pl_buf_destroy(gpu, &tbuf);
-            goto error;
+        for (int i = 0; i < num_slices; i++) {
+            // Restore buffer offset after downloading into temporary buffer,
+            // because we still need to copy the data from the temporary buffer
+            // into this offset in the original buffer
+            const size_t tmp_offset = slices[i].buf_offset;
+            slices[i].buf_offset = 0;
+            bool ok = fmt->emulated ? pl_tex_download_texel(gpu, &slices[i])
+                                    : pl_tex_download(gpu, &slices[i]);
+            slices[i].buf_offset = tmp_offset;
+            if (!ok)
+                goto error;
         }
 
+        // Finally, download into the user buffer
         struct vk_cmd *cmd = CMD_BEGIN_TIMED(tex_vk->transfer_queue, params->timer);
-        if (!cmd) {
-            pl_buf_destroy(gpu, &tbuf);
+        if (!cmd)
             goto error;
-        }
 
-        struct pl_buf_vk *tbuf_vk = PL_PRIV(tbuf);
-        VkBufferCopy region = {
-            .srcOffset = tbuf_vk->mem.offset,
-            .dstOffset = buf_offset,
-            .size = size,
-        };
-
-        vk_buf_barrier(gpu, cmd, tbuf, VK_PIPELINE_STAGE_2_COPY_BIT,
-                       VK_ACCESS_2_TRANSFER_READ_BIT, 0, size, false);
         vk_buf_barrier(gpu, cmd, buf, VK_PIPELINE_STAGE_2_COPY_BIT,
                        VK_ACCESS_2_TRANSFER_WRITE_BIT, params->buf_offset, size,
                        false);
-        vk->CmdCopyBuffer(cmd->buf, tbuf_vk->mem.buf, buf_vk->mem.buf,
-                          1, &region);
+
+        for (int i = 0; i < num_slices; i++) {
+            pl_buf slice = slices[i].buf;
+            struct pl_buf_vk *slice_vk = PL_PRIV(slice);
+            vk_buf_barrier(gpu, cmd, slice, VK_PIPELINE_STAGE_2_COPY_BIT,
+                           VK_ACCESS_2_TRANSFER_READ_BIT, 0, slice->params.size,
+                           false);
+
+            vk->CmdCopyBuffer(cmd->buf, slice_vk->mem.buf, buf_vk->mem.buf, 1, &(VkBufferCopy) {
+                .srcOffset = slice_vk->mem.offset,
+                .dstOffset = buf_vk->mem.offset + slices[i].buf_offset,
+                .size      = slice->params.size,
+            });
+
+            pl_buf_destroy(gpu, &slices[i].buf);
+        }
+
         vk_buf_flush(gpu, cmd, buf, params->buf_offset, size);
 
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-
-        CMD_FINISH(&cmd);
-        pl_buf_destroy(gpu, &tbuf);
+        pl_free(slices);
+        return CMD_FINISH(&cmd);
 
     } else {
 
-        pl_assert(fmt->texel_align == fmt->texel_size);
+        pl_assert(params->row_pitch % fmt->texel_size == 0);
+        pl_assert(params->depth_pitch % params->row_pitch == 0);
         const VkBufferImageCopy region = {
             .bufferOffset = buf_offset,
             .bufferRowLength = params->row_pitch / fmt->texel_size,
@@ -1046,13 +1106,15 @@ bool vk_tex_download(pl_gpu gpu, const struct pl_tex_transfer_params *params)
         if (params->callback)
             vk_cmd_callback(cmd, tex_xfer_cb, params->callback, params->priv);
 
-
-        CMD_FINISH(&cmd);
+        return CMD_FINISH(&cmd);
     }
 
-    return true;
+    pl_unreachable();
 
 error:
+    for (int i = 0; i < num_slices; i++)
+        pl_buf_destroy(gpu, &slices[i].buf);
+    pl_free(slices);
     return false;
 }
 
