@@ -52,8 +52,13 @@ struct vk_malloc_params {
     enum pl_handle_type export_handle;
     enum pl_handle_type import_handle;
     struct pl_shared_mem shared_mem; // for `import_handle`
+    pl_debug_tag debug_tag;
 };
 
+// Returns the amount of available memory matching a given set of property
+// flags. Always returns the highest single allocation, not the combined total.
+size_t vk_malloc_avail(struct vk_malloc *ma, VkMemoryPropertyFlags flags);
+
 bool vk_malloc_slice(struct vk_malloc *ma, struct vk_memslice *out,
                      const struct vk_malloc_params *params);
 
